package shared_workspace

import (
	"encoding/json"
	"slices"
	"strings"

	"cloud.o-forge.io/core/oc-lib/dbs"
	"cloud.o-forge.io/core/oc-lib/dbs/mongo"
	"cloud.o-forge.io/core/oc-lib/models/peer"
	"cloud.o-forge.io/core/oc-lib/models/utils"
	w "cloud.o-forge.io/core/oc-lib/models/workflow"
	"cloud.o-forge.io/core/oc-lib/models/workspace"
	"cloud.o-forge.io/core/oc-lib/tools"
)

type sharedWorkspaceMongoAccessor struct {
	utils.AbstractAccessor
}

func New() *sharedWorkspaceMongoAccessor {
	return &sharedWorkspaceMongoAccessor{}
}

func (wfa *sharedWorkspaceMongoAccessor) DeleteOne(id string) (utils.DBObject, int, error) {
	wfa.sharedWorkflow(&SharedWorkspace{}, id)
	wfa.sharedWorkspace(&SharedWorkspace{}, id)
	return wfa.GenericDeleteOne(id, wfa)
}

func (wfa *sharedWorkspaceMongoAccessor) sharedWorkspace(shared *SharedWorkspace, id string) {
	eldest, code, _ := wfa.LoadOne(id)
	accessor := (&workspace.Workspace{}).GetAccessor(nil)
	if code == 200 {
		eld := eldest.(*SharedWorkspace)
		if eld.Workspaces != nil {
			for _, v := range eld.Workspaces {
				accessor.UpdateOne(&workspace.Workspace{Shared: false}, v)
				if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.WORKSPACE.String()] == nil {
					continue
				}
				methods := wfa.Caller.URLS[utils.WORKSPACE.String()]
				if _, ok := methods[tools.DELETE]; !ok {
					continue
				}
				for _, p := range shared.Peers {
					pp, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(p)
					if code == 200 {
						resp, err := wfa.Caller.CallDelete(pp.(*peer.Peer).Url, strings.ReplaceAll(methods[tools.DELETE], ":id", v))
						if err != nil {
							wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error: " + err.Error())
						}
						var r map[string]interface{}
						json.Unmarshal(resp, &r)
						if e, ok := r["error"]; ok && e != "" {
							wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error" + e.(string))
						}
					}
				}
			}
		}
	}
	if shared.Workspaces != nil {
		for _, v := range shared.Workspaces {
			workspace, code, _ := accessor.UpdateOne(&workspace.Workspace{Shared: true}, v)
			if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.WORKSPACE.String()] == nil {
				continue
			}
			methods := wfa.Caller.URLS[utils.WORKSPACE.String()]
			if _, ok := methods[tools.POST]; !ok {
				continue
			}
			for _, p := range shared.Peers {
				if code != 200 {
					continue
				}
				pp, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(p)
				if code == 200 {
					resp, err := wfa.Caller.CallPost(pp.(*peer.Peer).Url, methods[tools.POST], workspace.Serialize())
					if err != nil {
						wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error: " + err.Error())
					}
					var r map[string]interface{}
					json.Unmarshal(resp, &r)
					if e, ok := r["error"]; ok && e != "" {
						wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error" + e.(string))
					}
				}
			}
		}
	}
}

func (wfa *sharedWorkspaceMongoAccessor) sharedWorkflow(shared *SharedWorkspace, id string) {
	accessor := (&w.Workflow{}).GetAccessor(nil)
	eldest, code, _ := wfa.LoadOne(id)
	if code == 200 {
		eld := eldest.(*SharedWorkspace)
		if eld.Workflows != nil {
			for _, v := range eld.Workflows {
				data, code, _ := accessor.LoadOne(v)
				if code == 200 {
					s := data.(*w.Workflow)
					new := []string{}
					for _, id2 := range s.Shared {
						if id2 != id {
							new = append(new, id2)
						}
					}
					n := &w.Workflow{}
					n.Shared = new
					accessor.UpdateOne(n, v)
					if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.WORKFLOW.String()] == nil {
						continue
					}
					methods := wfa.Caller.URLS[utils.WORKFLOW.String()]
					if _, ok := methods[tools.DELETE]; !ok {
						continue
					}
					for _, p := range shared.Peers {
						pp, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(p)
						if code == 200 {
							resp, err := wfa.Caller.CallDelete(pp.(*peer.Peer).Url, strings.ReplaceAll(methods[tools.DELETE], ":id", v))
							if err != nil {
								wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error: " + err.Error())
							}
							var r map[string]interface{}
							json.Unmarshal(resp, &r)
							if e, ok := r["error"]; ok && e != "" {
								wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error" + e.(string))
							}
						}
					}
				}
			}
		}
	}
	if shared.Workflows != nil {
		for _, v := range shared.Workflows {
			data, code, _ := accessor.LoadOne(v)
			if code == 200 {
				s := data.(*w.Workflow)
				if !slices.Contains(s.Shared, id) {
					s.Shared = append(s.Shared, id)
					workflow, code, _ := accessor.UpdateOne(s, v)
					if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.WORKFLOW.String()] == nil {
						continue
					}
					methods := wfa.Caller.URLS[utils.WORKFLOW.String()]
					if _, ok := methods[tools.POST]; !ok {
						continue
					}
					for _, p := range shared.Peers {
						if code != 200 {
							continue
						}
						pp, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(p)
						if code == 200 {
							resp, err := wfa.Caller.CallPost(pp.(*peer.Peer).Url, methods[tools.POST], workflow.Serialize())
							if err != nil {
								wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error: " + err.Error())
							}
							var r map[string]interface{}
							json.Unmarshal(resp, &r)
							if e, ok := r["error"]; ok && e != "" {
								wfa.Logger.Error().Msg("Could not send to peer " + pp.(*peer.Peer).Url + ". Error" + e.(string))
							}
						}
					}
				}
			}
		}
	}
}

func (wfa *sharedWorkspaceMongoAccessor) deleteToPeer(shared *SharedWorkspace) {
	if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.SHARED_WORKSPACE.String()] == nil {
		return
	}
	methods := wfa.Caller.URLS[utils.SHARED_WORKSPACE.String()]
	if _, ok := methods[tools.DELETE]; !ok {
		return
	}
	for _, v := range shared.Peers {
		p, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(v)
		if code != 200 {
			continue
		}
		resp, err := wfa.Caller.CallDelete(p.(*peer.Peer).Url, strings.ReplaceAll(methods[tools.DELETE], ":id", shared.GetID()))
		if err != nil {
			wfa.Logger.Error().Msg("Could not send to peer " + p.(*peer.Peer).Url + ". Error: " + err.Error())
		}
		var r map[string]interface{}
		json.Unmarshal(resp, &r)
		if e, ok := r["error"]; ok && e != "" {
			wfa.Logger.Error().Msg("Could not send to peer " + p.(*peer.Peer).Url + ". Error" + e.(string))
		}
	}
}

func (wfa *sharedWorkspaceMongoAccessor) sendToPeer(shared *SharedWorkspace) {
	if wfa.Caller != nil || wfa.Caller.URLS == nil || wfa.Caller.URLS[utils.SHARED_WORKSPACE.String()] == nil {
		return
	}
	methods := wfa.Caller.URLS[utils.SHARED_WORKSPACE.String()]
	if _, ok := methods[tools.POST]; !ok {
		return
	}
	for _, v := range shared.Peers {
		p, code, _ := (&peer.Peer{}).GetAccessor(nil).LoadOne(v)
		if code != 200 {
			continue
		}
		resp, err := wfa.Caller.CallPost(p.(*peer.Peer).Url, methods[tools.POST], shared.Serialize())
		if err != nil {
			wfa.Logger.Error().Msg("Could not send to peer " + p.(*peer.Peer).Url + ". Error: " + err.Error())
		}
		var r map[string]interface{}
		json.Unmarshal(resp, &r)
		if e, ok := r["error"]; ok && e != "" {
			wfa.Logger.Error().Msg("Could not send to peer " + p.(*peer.Peer).Url + ". Error" + e.(string))
		}
	}
}

func (wfa *sharedWorkspaceMongoAccessor) UpdateOne(set utils.DBObject, id string) (utils.DBObject, int, error) {
	wfa.deleteToPeer(set.(*SharedWorkspace))
	wfa.sharedWorkflow(set.(*SharedWorkspace), id)
	wfa.sharedWorkspace(set.(*SharedWorkspace), id)
	wfa.sendToPeer(set.(*SharedWorkspace))
	return wfa.GenericUpdateOne(set.(*SharedWorkspace), id, wfa, &SharedWorkspace{})
}

func (wfa *sharedWorkspaceMongoAccessor) StoreOne(data utils.DBObject) (utils.DBObject, int, error) {
	wfa.deleteToPeer(data.(*SharedWorkspace))
	d, code, err := wfa.GenericStoreOne(data.(*SharedWorkspace), wfa)
	if code == 200 {
		wfa.sharedWorkflow(d.(*SharedWorkspace), d.GetID())
		wfa.sharedWorkspace(d.(*SharedWorkspace), d.GetID())
		wfa.sendToPeer(d.(*SharedWorkspace))
	}
	return data, code, err
}

func (wfa *sharedWorkspaceMongoAccessor) CopyOne(data utils.DBObject) (utils.DBObject, int, error) {
	return wfa.StoreOne(data)
}

func (wfa *sharedWorkspaceMongoAccessor) LoadOne(id string) (utils.DBObject, int, error) {
	var sharedWorkspace SharedWorkspace
	res_mongo, code, err := mongo.MONGOService.LoadOne(id, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve " + id + " from db. Error: " + err.Error())
		return nil, code, err
	}
	res_mongo.Decode(&sharedWorkspace)

	return &sharedWorkspace, 200, nil
}

func (wfa sharedWorkspaceMongoAccessor) LoadAll() ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	res_mongo, code, err := mongo.MONGOService.LoadAll(wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not retrieve any from db. Error: " + err.Error())
		return nil, code, err
	}
	var results []SharedWorkspace
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, &r)
	}
	return objs, 200, nil
}

func (wfa *sharedWorkspaceMongoAccessor) Search(filters *dbs.Filters, search string) ([]utils.ShallowDBObject, int, error) {
	objs := []utils.ShallowDBObject{}
	if (filters == nil || len(filters.And) == 0 || len(filters.Or) == 0) && search != "" {
		filters = &dbs.Filters{
			Or: map[string][]dbs.Filter{
				"abstractobject.name": {{Operator: dbs.LIKE.String(), Value: search}},
			},
		}
	}
	res_mongo, code, err := mongo.MONGOService.Search(filters, wfa.GetType())
	if err != nil {
		wfa.Logger.Error().Msg("Could not store to db. Error: " + err.Error())
		return nil, code, err
	}
	var results []SharedWorkspace
	if err = res_mongo.All(mongo.MngoCtx, &results); err != nil {
		return nil, 404, err
	}
	for _, r := range results {
		objs = append(objs, &r)
	}
	return objs, 200, nil
}
